
import sys
import DAO as dao 

class StateNode():
    def __init__(self, num, term):
        self.edges = dict()
        self.number = num
        self.name = "State " + str(self.number)
        self.default = None
        self.visited = False
        self.tokens = []
        self.errors = term.copy()
        self.errorSets = []
        self.accept = []
        
    def __str__(self) -> str:
        return self.name

STATEGRAPH = dict()
START = 4
STATE = 5
TERMINALS = []
NONTERMINALS = []
SUBSTITUTES = dict()


def getEqTokens(line, token):
    token = token.strip()
    line = line.strip()
    start = 0
    try:
        line.index(' ')
    except:
        if TERMINALS.count(token) == 0 and TERMINALS.count(line) == 0:
            return {token, line}      

def combineEquivalents(Equivalents):
    tokenSets = []
    for pair in Equivalents:
        for tokenSet in tokenSets:
            if not tokenSet.isdisjoint(pair):
                pair = tokenSet.union(pair)
                tokenSets.remove(tokenSet)
                continue
        tokenSets.append(set(pair))
    if len(tokenSets) > 0:
        for thisSet in tokenSets:
            for thatSet in tokenSets:
                if thisSet != thatSet:
                    for token in thisSet:
                        if not thatSet.isdisjoint(token):
                            thatSet.add(thisSet)
                            tokenSets.remove(thisSet)
                            break
    return tokenSets
    
def grammarAnalysis(lines):
    Equivalents = []
    grammar = False
    token = ''
    for line in lines:
        line = line.strip()
        if 'Grammar' in line:
            grammar = True
            continue
        if line == '':
            continue
        if grammar:
            count = 0
            if ':' in line:
                start = line.index(' ') + 1
                end = line.index(':')
                token = line[start:end]
                line = line[end + 1:]
                result = getEqTokens(line, token)
                if result != None:
                    Equivalents.append(result)
            if '|' in line:
                start = line.index('|') + 2
                line = line[start:]
                result = getEqTokens(line, token)
                if result != None:
                    Equivalents.append(result)
            if 'Terminals' in line or 'Nonterminals' in line:
                grammar = False
                break
    return combineEquivalents(Equivalents)

def getTerminals(lines):
    terms = False
    for line in lines:
        if 'Terminals' in line:
            terms = True
            continue
        if terms:
            if len(line) > 1 and 'Nonterminals' not in line:
                stop = line[START:].index(' ') + START
                TERMINALS.append(line[START:stop])
            if 'Nonterminals' in line:
                break

def getNonTerminals(lines):
    terms = False
    for line in lines:
        if 'Nonterminals' in line:
            terms = True
            continue
        if terms:
            if len(line) > 1 and 'State' not in line and line[START] != ' ':
                try:
                    stop = line[START:].index(' ') + START
                except:
                    stop = len(line) - 1
                NONTERMINALS.append(line[START:stop])
            if 'State' in line or 'Grammar' in line:
                break
            

def makeStates(lines):
    for line in lines:
        if line[:STATE] == 'State':
            st = int(line[STATE + 1:])
            STATEGRAPH[st] = StateNode(st, TERMINALS)
            #print(STATEGRAPH[st])

def makeEdges(lines):
    state = -1
    for line in lines:
        if line[:STATE] == 'State':
            state = int(line[STATE + 1:])
            continue
        if state >= 0:
            if len(line) > 1 and not line[START].isdigit():
                stop = line[START:].index(' ') + START
                word = line[START:stop]
                if word == '$default':
                    if line[-2:-1] == ")":
                        start = line.index('(') + 1
                        stop = line.index(')')
                        STATEGRAPH[state].default = line[start:stop]
                    else:
                        STATEGRAPH[state].default = line[14:]
                    STATEGRAPH[state].errors.clear()
                elif word in TERMINALS or word in NONTERMINALS:
                    # ignore special cases for now
                    if '[' in line and ']' in line:  # a line with a terminal or non-terminal followed by square brackets are ignored
                        continue
                    STATEGRAPH[state].accept += [word]
                    pos = line.index('go to state')
                    st = int(line[pos + 11:])
                    STATEGRAPH[state].edges[word] = STATEGRAPH[st]
                    if STATEGRAPH[state].errors != [] and word in TERMINALS:
                        STATEGRAPH[state].errors.remove(word)

def graphTraversal(current, t = []):
    tokens = []
    if not current.visited:
        if t != []:
            tokens = t
        current.tokens = [tokens]
        current.visited = True
        for token in current.edges.keys():
            graphTraversal(current.edges[token], tokens + [token])
    else:          
        if t != []:
            current.tokens.append(t)

def snippetTest(errorSet, index1, sub1, index2, sub2):
    stack = [STATEGRAPH[0]]
    i = 0
    for token in errorSet:
        if i == index1:
            token = sub1
        if i == index2:
            token = sub2
        while len(stack) > 0:
            edgeTo = None
            state = stack[len(stack) - 1]
            try:
                edgeTo = state.edges[token]
            except KeyError:
#                print(str(errorSet) + " : True")
                return True
            if edgeTo != None: 
                stack.append(edgeTo)
                i += 1
#                print(str(errorSet) + " : Break")
                break
#    print(str(errorSet) + " : False")
    return False


def makeErrorSnippets(equivalents):
#    print("Equivalents: " + str(equivalents))
    count = 0
    for key in STATEGRAPH.keys():
#        print(key)
        for tokens in STATEGRAPH[key].tokens:
#            print(tokens)
            swap1 = len(tokens) - 1
            swap2 = len(tokens)
            for error in STATEGRAPH[key].errors:
                #print(error)
                errorSet = tokens + [error]
                for eqSet1 in equivalents: #eqSet1 Loop Begin
                    for eqSet2 in equivalents: #eqSet2 Loop Begin
                        success = True
#                        print('got here')
                        if errorSet[swap1] in eqSet1: #swap1 token is in a equivalent set
                            if errorSet[swap2] in eqSet2: #swap2 token is in a equivalent set
                                for token1 in eqSet1:
                                    for token2 in eqSet2:
                                        result = snippetTest(errorSet, swap1, token1, swap2, token2)
                                        if result == False:
                                            success = False
                            else: #swap2 token is not in a equivalent set
                                token2 = errorSet[swap2]
                                for token1 in eqSet1:
                                    result = snippetTest(errorSet, swap1, token1, swap2, token2)
                                    if result == False:
                                        success = False
                        else: #swap1 token is not in a equivalent set
                            token1 = errorSet[swap1]
                            if errorSet[swap2] in eqSet2: #swap2 token is in a equivalent set
                                for token2 in eqSet2:
                                    result = snippetTest(errorSet, swap1, token1, swap2, token2)
                                    if result == False:
                                        success = False
                            else: #swap2 token is not in a equivalent set
                                token2 = errorSet[swap2]
                                result = snippetTest(errorSet, swap1, token1, swap2, token2)
                                if result == False:
                                    success = False   
                if success:
                    count += 1
                    STATEGRAPH[key].errorSets += [errorSet]
                    #eqSet2 Loop End
    return count

def makeMessage(st, tp):
    message = "\n\t::: *** syntax error ***"
    match tp:
        case 'options':
            message = '\n::: The following would be acceptable: '
            count = 0
            for option in STATEGRAPH[st].accept:
                if option in TERMINALS:
                    if count > 0:
                        message += ', '
                    message += option
                    count += 1
        case 'where':
            pass

    return message

def makeMerrFile(tp):
    merrFile = dao.WriteFile("merr.txt")
    for key in STATEGRAPH.keys():
    #    merrFile.writeLine(STATEGRAPH[key].name)
        if STATEGRAPH[key].errors != []:
            for errorSet in STATEGRAPH[key].errorSets:
                merr = ""
                for token in errorSet:
                    if len(SUBSTITUTES) > 0:
                        try:
                            merr += SUBSTITUTES[token] + " "
                        except:
                            merr += token + " "
                    else:
                        merr += token + " "
                merr = merr + makeMessage(key, tp)
                merrFile.writeLine(merr)
    merrFile.close()

def printGraph():
    for key in STATEGRAPH.keys():
        print("\n" + STATEGRAPH[key].name)
        for token in STATEGRAPH[key].edges.keys():
            print("\t" + token + " => " + STATEGRAPH[key].edges[token].name)
        if STATEGRAPH[key].default != None:
            print("\t$default <- " + STATEGRAPH[key].default)
        if STATEGRAPH[key].errors != []:
            print("\tError Sets:")
            for errorSet in STATEGRAPH[key].errorSets:
                print("\t\t" + str(errorSet))
        print("\tTraversal: " + str(STATEGRAPH[key].tokens))
    
def printlist(L):
    for item in L:
        print("\t" + item)

def main():
    infile = dao.ReadFile(sys.argv[1])
    lines = infile.readlines()
    infile.close()
    if len(sys.argv) > 2:
        infile = dao.ReadFile(sys.argv[2])
        line = infile.readline()
        while line != "":
            index = line.find(":")
            symbol = line[:index].strip()
            sub = line[index + 1:].strip()
            SUBSTITUTES[symbol] = sub
            line = infile.readline()
        
    getTerminals(lines)
    getNonTerminals(lines)
    equivalents = grammarAnalysis(lines)
    makeStates(lines)
    makeEdges(lines)
    graphTraversal(STATEGRAPH[0])
    count = makeErrorSnippets(equivalents)

    tp = 'options'
    print('TERMINALS:')
    printlist(TERMINALS)
    print('NONTERMINALS:')
    printlist(NONTERMINALS)
    print('EQUIVALENTS:')
    print(str(equivalents))
    makeMerrFile(tp)
    printGraph()
    print(str(count) + " Error Snippets")
    
main()